package quimufu.colourful_portals.portal;

import net.minecraft.entity.Entity;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.Identifier;
import net.minecraft.util.math.BlockBox;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Direction;
import net.minecraft.util.math.Vec3d;
import net.minecraft.world.TeleportTarget;
import quimufu.colourful_portals.general_util.LinkedList;
import quimufu.colourful_portals.general_util.Node;

import java.util.*;

import static quimufu.colourful_portals.ColourfulPortalsMod.LOGGER;
import static quimufu.colourful_portals.ColourfulPortalsMod.MOD_ID;
import static quimufu.colourful_portals.portal.PortalHelper.getAxisW;
import static quimufu.colourful_portals.portal.PortalHelper.insideOf;
import static quimufu.colourful_portals.portal.PortalHelper.getDimId;

public class DefaultLinkingSystem implements PortalLinkingSystem {

    public static final Identifier DEFAULT_LINKING_SYSTEM = Identifier.of(MOD_ID, "default_linking_system");

    private final Map<BlockPos, Set<Node<PortalRepresentation>>> positionalLookup = new HashMap<>();
    private final MinecraftServer server;

    public DefaultLinkingSystem(MinecraftServer minecraftServer) {
        this.server = minecraftServer;
    }

    @Override
    public Identifier getLinkingSystemId() {
        return DEFAULT_LINKING_SYSTEM;
    }

    @Override
    public void linkPortals(LinkedList<PortalRepresentation> portalRepresentations) {
        LOGGER.info("start linkPortals");
        for (Node<PortalRepresentation> node = portalRepresentations.getNode(0); node != null; node = node.getNext()) {
            PortalRepresentation portalRepresentation = node.getValue();

            Node<PortalRepresentation> finalNode = node;
            insideOf(portalRepresentation)
                    .forEachRemaining(blockPos -> {
                        Set<Node<PortalRepresentation>> portalsAtPosition = positionalLookup.computeIfAbsent(blockPos, (k) -> new HashSet<>(1));
                        portalsAtPosition.add(finalNode);
                    });
        }
        LOGGER.info("end linkPortals");
    }

    @Override
    public void unLinkPortal(PortalRepresentation portalRepresentation) {
        //no-op
    }

    @Override
    public void onPortalPassed(Entity entity, BlockPos pos, ServerWorld world, Direction.Axis a) {
        Set<Node<PortalRepresentation>> portalRepresentationCandidated = positionalLookup.get(pos);
        Optional<Node<PortalRepresentation>> portalOpt = portalRepresentationCandidated.stream()
                .filter(n -> !n.orphaned())
                .filter(n -> n.getValue() != null)
                .filter(n -> getAxisW(n.getValue()).rotateYClockwise().getAxis() == a)
                .filter(n -> Objects.equals(n.getValue().dimensionId(), getDimId(world)))
                .findAny();
        if (portalOpt.isPresent()) {
            Node<PortalRepresentation> fromNode = portalOpt.get();
            Node<PortalRepresentation> toNode = fromNode.getNext() != null ? fromNode.getNext() : fromNode.getPartOf().getNode(0);
            if (toNode == null || toNode.getValue() == null) {
                return;
            }

            PortalRepresentation fromPortal = fromNode.getValue();
            PortalRepresentation toPortal = toNode.getValue();
            BlockBox fromBox = fromPortal.location();
            Vec3d fromCenter = new Vec3d((fromBox.getMaxX()+1 + fromBox.getMinX())/2D,(fromBox.getMaxY()+1 + fromBox.getMinY())/2D,(fromBox.getMaxZ()+1 + fromBox.getMinZ())/2D);
            BlockBox toBox = toPortal.location();
            Vec3d toCenter = new Vec3d((toBox.getMaxX()+1 + toBox.getMinX())/2D,(toBox.getMaxY()+1 + toBox.getMinY())/2D,(toBox.getMaxZ()+1 + toBox.getMinZ())/2D);
            if (getAxisW(fromPortal) == getAxisW(toPortal)) {
                Vec3d relPos = entity.getPos().subtract(fromCenter);
                Vec3d targetPos = toCenter.add(relPos);
                //todo: maybe fancy continoous movement math!
                ServerWorld toWorld = getPortalWorld(server, toPortal);
                TeleportTarget teleportTarget = new TeleportTarget(toWorld, targetPos, entity.getVelocity(), entity.getYaw(), entity.getPitch(), TeleportTarget.ADD_PORTAL_CHUNK_TICKET);
                entity.teleportTo(teleportTarget);
            } else {
                Vec3d relPos = entity.getPos().subtract(fromCenter);
                Vec3d targetPos = toCenter.add(new Vec3d(-relPos.z, relPos.y, relPos.x));
                //todo: maybe fancy continoous movement math!
                ServerWorld toWorld = getPortalWorld(server, toPortal);
                TeleportTarget teleportTarget = new TeleportTarget(toWorld, targetPos, entity.getVelocity(), (entity.getYaw() + 90.F) % 360.F, entity.getPitch(), TeleportTarget.ADD_PORTAL_CHUNK_TICKET);
                entity.teleportTo(teleportTarget);
            }
        }
    }

    @Override
    public boolean needsReInit() {
        return true;
    }

    private static ServerWorld getPortalWorld(MinecraftServer server, PortalRepresentation fromPortalRepresentation) {
        ServerWorld serverWorld = PortalHelper.getPortalWorld(server, fromPortalRepresentation);
        if (serverWorld == null) {
            LOGGER.warn("couldn't get portal dimensionId for portal {}. Don't sue me!", fromPortalRepresentation);
            throw new RuntimeException();
        }
        return serverWorld;
    }
}
